// Copyright 2011 Google Inc. All Rights Reserved.

/**
 * @fileoverview A find in page tool.  It scans the DOM for elements with the
 * text being search for, and wraps them with a span that highlights them.
 *
 * @author bmcmahan@google.com (Benjamin McMahan)
 */

/**
 * Namespace for this file.
 */
var findInPage = {};

/**
 * Index of the current highlighted choice.  -1 means none.
 * @type {number}
 */
findInPage.index = -1;

/**
 * The list of found searches in span form.
 * @type {Array.<Element>}
 */
findInPage.spans = [];

/**
 * Associate array to stash element styles while the element is highlighted.
 * @type {Object.<Element,Object<string,string>>}
 */
findInPage.savedElementStyles = {};

/**
 * The style DOM element that we add.
 * @type {Element}
 */
findInPage.style = null;

/**
 * Width we expect the page to be.  For example (320/480) for iphone,
 * (1024/768) for ipad.
 * @type {number}
 */
findInPage.pageWidth = 320;

/**
 * Height we expect the page to be.
 * @type {number}
 */
findInPage.pageHeight = 480;

/**
 * Node names that are not going to be processed.
 * @type {Object}
 */
findInPage.ignoreNodeNames = {
 'SCRIPT': 1,
 'STYLE': 1,
 'EMBED': 1,
 'OBJECT': 1
};

findInPage.CSS_CLASS_NAME = 'find_in_page';

/**
 * Maximum zoom scale for selected elements.
 * @type {Number}
 */
findInPage.MAX_SELECTION_ZOOM = 1.5;

/**
 * Result passed back to app to indicate no results for the query.
 * @type {string}
 */
findInPage.NO_RESULTS = '[0,[0,0]]';

/**
 * Regex to escape regex special characters in a string.
 * @type {RegExp}
 */
findInPage.REGEX_ESCAPER = /([.?*+^$[\]\\(){}|-])/g;

/**
 * Creates the regex needed to find the text.
 * @param {string} findText Phrase to look for.
 * @param {boolean} opt_split True to split up the phrase.
 * @return {RegExp}
 */
findInPage.getRegex = function(findText, opt_split) {
  var regexString = '';
  if (opt_split) {
    var words = [];
    var split = findText.split(' ');
    for (var i = 0; i < split.length; i++) {
      words.push(findInPage.escapeRegex(split[i]));
    }
    var joinedWords = words.join('|');
    regexString = '(' +
        // Match at least one word.
        '\\b(?:' + joinedWords + ')' +
        // Include zero or more additional words separated by whitespace.
        '(?:\\s*\\b(?:' + joinedWords + '))*' +
        ')';
  } else {
    regexString = '(' + findInPage.escapeRegex(findText) + ')';
  }
  return new RegExp(regexString, 'ig');
};

/**
 * Looks for a phrase in the DOM.
 * @param {string} findText Phrase to look for like "ben franklin".
 * @param {boolean} opt_split True to split up the words and look for any
 *     of them.  False to require the full phrase to be there.
 *     Undefined will try the full phrase, and if nothing is found do the split.
 * @return {number} How many results there are in the page.
 */
findInPage.highlightWord = function(findText, opt_split) {
  if (!findText || !findText.replace(/\u00a0|\s/g,'')) {
    // No searching for emptyness.
    return findInPage.NO_RESULTS;
  }
  if (findInPage.spans && findInPage.spans.length) {
    // Clean up a previous run.
    findInPage.clearHighlight();
  }
  // Store all DOM modifications to do them in a tight loop at once.
  var replacements = [];
  // Node is what we are currently looking at.
  var node = document.body;
  // Holds what nodes we have not processed yet.
  var stack = [];
  var regex = findInPage.getRegex(findText, opt_split);
  // Go through every node in DFS fashion.
  while (node) {
    // Add all children that are visible, non-zero sized DOM elements or text.
    var children = node.childNodes;
    if (children && children.length) {
      for (var i = children.length - 1; i >= 0; --i) {
        var child = children[i];
        if (((child.nodeType == Node.ELEMENT_NODE &&
              child.getBoundingClientRect().width != 0 &&
              child.getBoundingClientRect().height != 0) ||
             child.nodeType == Node.TEXT_NODE) &&
            !child.hidden &&
            !findInPage.ignoreNodeNames[child.nodeName]) {
          stack.push(children[i]);
        }
      }
    }

    // Find all matches in text nodes.
    if (node.nodeType == 3 && node.parentNode) {
      var strIndex = 0;
      var nodes = [];
      var match;
      while (match = regex.exec(node.textContent)) {
        try {
          var matchText = match[0];

          // If there is content before this match, add it to a new text node.
          if (match.index > 0) {
            var nodeSubstr = node.textContent.substring(strIndex,
                                                        match.index);
            nodes.push(document.createTextNode(nodeSubstr));
          }

          // Now create our matched span.
          var span = document.createElement('span');
          span.setAttribute('class', findInPage.CSS_CLASS_NAME);
          span.innerHTML = findInPage.escapeHTML(matchText);
          nodes.push(span);

          strIndex = match.index + matchText.length;
        } catch (e) {
          // Do nothing.
        }
      }
      if (nodes.length) {
        // Add any text after our matches to a new text node.
        if (strIndex < node.textContent.length) {
          var substr = node.textContent.substring(strIndex,
                                                  node.textContent.length);
          nodes.push(document.createTextNode(substr))
        }
        replacements.push({oldNode: node, newNodes: nodes});
        regex.lastIndex = 0;
      }
    }

    if (stack.length > 0) {
      node = stack.pop();
    } else {
      node = null;
    }
  }
  // Insert each of the replacement nodes into the old node's parent, then
  // remove the old node.
  replacements.forEach(function (replacement) {
    var parent = replacement.oldNode.parentNode;
    replacement.newNodes.forEach(function (node) {
      parent.insertBefore(node, replacement.oldNode);
    });
    parent.removeChild(replacement.oldNode);
  });
  findInPage.spans = document.getElementsByClassName(findInPage.CSS_CLASS_NAME);
  var pos;
  // Try again flow.
  // If opt_split is true, we are done since we won't do any better.
  // If opt_split is false, they were explicit about wanting the full thing
  //    so do not try with a split.
  // If opt_split is undefined and we did not find an answer, go ahead and try
  //    splitting the terms.
  if (findInPage.spans.length == 0 && opt_split === undefined) {
    // Try to be more aggressive:
    return findInPage.highlightWord(findText, true);
  } else {
    pos = findInPage.goNext();
    if (pos) {
      return '[' + findInPage.spans.length + ',' + pos + ']';
    } else if (opt_split === undefined) {
      // Nothing visible, go ahead and be more aggressive.
      return findInPage.highlightWord(findText, true);
    } else {
      return findInPage.NO_RESULTS;
    }
  }
};

/**
 * Removes all currently highlighted spans.
 * Note: It does not restore previous state, just removes the class name.
 */
findInPage.clearHighlight = function() {
  if (findInPage.index >= 0) {
    findInPage.removeSelectHighlight(findInPage.spans[findInPage.index]);
  }
  // Store all DOM modifications to do them in a tight loop.
  var modifications = [];
  var length = findInPage.spans.length;
  var prevParent = null;
  for (var i = length - 1; i >= 0; i--) {
    var elem = findInPage.spans[i];
    var parentNode = elem.parentNode;
    var nodeText = elem.innerText;
    // If this element has the same parent as the previous, check if we should
    // add this node to the previous one.
    if (prevParent && prevParent.isSameNode(parentNode) &&
        elem.nextSibling.isSameNode(findInPage.spans[i + 1].previousSibling)) {
      var prevMod = modifications[modifications.length - 1];
      prevMod.nodesToRemove.push(elem);
      var elemText = elem.innerText;
      if (elem.previousSibling) {
        prevMod.nodesToRemove.push(elem.previousSibling);
        elemText = elem.previousSibling.textContent + elemText;
      }
      prevMod.replacement.textContent =
          elemText + prevMod.replacement.textContent;
    }
    else { // Element isn't attached to previous, so create a new modification.
      var nodesToRemove = [elem];
      if (elem.previousSibling && elem.previousSibling.nodeType == 3) {
        nodesToRemove.push(elem.previousSibling);
        nodeText = elem.previousSibling.textContent + nodeText;
      }
      if (elem.nextSibling && elem.nextSibling.nodeType == 3) {
        nodesToRemove.push(elem.nextSibling);
        nodeText = nodeText + elem.nextSibling.textContent;
      }
      var textNode = document.createTextNode(nodeText);
      modifications.push({nodesToRemove: nodesToRemove, replacement: textNode});
    }
    prevParent = parentNode;
  }
  var numMods = modifications.length;
  for (i = numMods - 1; i >= 0; i--) {
    var mod = modifications[i];
    for (var j = 0; j < mod.nodesToRemove.length; j++) {
      var existing = mod.nodesToRemove[j];
      if (j == 0) {
        existing.parentNode.replaceChild(mod.replacement, existing);
      } else {
        existing.parentNode.removeChild(existing);
      }
    }
  }

  findInPage.spans = [];
  findInPage.index = -1;
};

/**
 * Switches to the subsequent result, animating a little highlight in the
 * process.
 * @param {boolean} isForward Find the next or previous result.
 * @return {string} JSON encoded array of coordinates to scroll to, or blank if
 *     nothing happened.
 */
findInPage.subsequentResult = function(isForward) {
  if (!findInPage.spans || findInPage.spans.length == 0) {
    return '';
  }
  if (findInPage.index >= 0) {
    // Remove previous highlight.
    findInPage.removeSelectHighlight(findInPage.spans[findInPage.index]);
  }

  /**
   * Advances findInPage.index, wrapping around either end if reached.
   * @return {boolean} Returns true if wrapped around the end.
   */
  var advanceIndex = function() {
    if (isForward) {
      findInPage.index++;
      if (findInPage.index >= findInPage.spans.length) {
        findInPage.index = 0;
        return true;
      }
    } else {
      findInPage.index--;
      if (findInPage.index < 0) {
        findInPage.index = findInPage.spans.length - 1;
        return true;
      }
    }
    return false;
  };

  // Iterate through to the next index, but because they might not be visible,
  // keep trying until you find one that is.  Make sure we don't loop forever by
  // stopping on what we are currently highlighting.
  var old = findInPage.index;
  advanceIndex();

  while (old != findInPage.index &&
      !findInPage.isVisible(findInPage.spans[findInPage.index])) {
    if (advanceIndex() && old < 0) {
      // Didn't find anything and haven't highlighted anything yet.
      return '';
    }
  }
  // Return scroll dimensions.
  return findInPage.findScrollDimensions();

};

/**
 * Switches to the next result, animating a little highlight in the process.
 * @return {string} JSON encoded array of coordinates to scroll to, or blank if
 *     nothing happened.
 */
findInPage.goNext = function() {
  return findInPage.subsequentResult(true);
};

/**
 * Switches to the previous result, animating a little highlight in the process.
 * @return {string} JSON encoded array of coordinates to scroll to, or blank if
 *     nothing happened.
 */
findInPage.goPrev = function() {
  return findInPage.subsequentResult(false);
};

/**
 * Adds the special highlighting to the result at the index.
 * @param {number} opt_index Index to replace findInPage.index with.
 */
findInPage.addHighlightToIndex = function(opt_index) {
  if (opt_index !== undefined) {
    findInPage.index = opt_index;
  }
  findInPage.addSelectHighlight(findInPage.spans[findInPage.index]);
};

/**
 * Updates the elements style, while saving the old style in
 * findInPage.savedElementStyles.
 * @param {Element} element Element to update.
 * @param {string} style Name of the style to update.
 * @param {string} value New style value.
 */
findInPage.updateElementStyle = function(element, style, value) {
  if (!findInPage.savedElementStyles[element]) {
    findInPage.savedElementStyles[element] = {};
  }
  // We need to keep the original style setting for this element, so if we've
  // already saved a value for this style don't update it.
  if (!findInPage.savedElementStyles[element][style]) {
    findInPage.savedElementStyles[element][style] = element.style[style];
  }
  element.style[style] = value;
};

/**
 * Adds selected highlight style to the specified element.
 * @param {Element} element Element to highlight.
 */
findInPage.addSelectHighlight = function(element) {
  // Make sure this element will fit on the screen after we zoom.
  var bodyWidth = findInPage.getBodyWidth();
  var windowWidth = window.innerWidth;
  var zoom = findInPage.getHighlightZoomFactor();

  var origOffsetWidth = element.offsetWidth;
  // The findysel class we're about to add is already inline-block, but this
  // change can update the flow of the document (for example, it forces a
  // linefeed if the element was wrapped, or it could shift below an element
  // floated to the side).
  findInPage.updateElementStyle(element, 'display', 'inline-block');

  // inline-block elements will not flow around adjacent floated blocks. If it
  // is too wide to fit next to a floating element, it will reflow below the
  // floated element, which we don't want to happen. When this happens the
  // offsetWidth increases; constrain the element to the old width so it will
  // fit.
  if (element.offsetWidth > origOffsetWidth) {
    findInPage.updateElementStyle(element, 'width', origOffsetWidth + 'px');
  }

  // Update these variables now that the display has changed.
  var elementOffset = findInPage.findAbsolutePosition(element)[0];
  var zoomedWidth = element.offsetWidth * zoom +
      findInPage.MAX_HIGHLIGHT_PADDING * 2;

  if (elementOffset + zoomedWidth > windowWidth) {
    if (zoomedWidth > windowWidth) {
      // If the element doesn't fit even without the offset, constrain it to a
      // width that will.

      // We can't set offsetWidth directly so get the padding on our element.
      var elementWidth = window.getComputedStyle(element)
          .getPropertyCSSValue('width')
          .getFloatValue(CSSPrimitiveValue.CSS_PX);
      var elementPadding = (element.offsetWidth - elementWidth);
      var constrainedWidth = (windowWidth - elementOffset - elementPadding -
          findInPage.MAX_HIGHLIGHT_PADDING * 2) / zoom;

      // Now update the element width so it fits.
      findInPage.updateElementStyle(element, 'width', constrainedWidth + 'px');
      // Update zoomedWidth to reflect this change.
      zoomedWidth = element.offsetWidth * zoom +
          findInPage.MAX_HIGHLIGHT_PADDING * 2;
    }
    // Now the element will fit on the page, but it may be offset too much.
    var overflow = (elementOffset + zoomedWidth) - bodyWidth;
    if (overflow > 0) {
      // Override the transform to shift to the left when we zoom.
      var overflowRatio = overflow / (zoomedWidth - element.offsetWidth);
      findInPage.updateElementStyle(element,
          'webkitTransformOrigin',
          (overflowRatio * 100) + '% 50%');
      findInPage.updateElementStyle(element,
          'webkitTransform',
          'scale(' + zoom + ')');
    }
  }
  element.className = (element.className || '') + ' findysel';
};

/**
 * Removes selected highlight style from the specified element.
 * @param {Element} element Element to remove highlighting from.
 */
findInPage.removeSelectHighlight = function(element) {
  element.className = (element.className || '').replace(/\sfindysel/g, '');

  // Restore any styles we may have saved when adding the select highlighting.
  var savedStyles = findInPage.savedElementStyles[element];
  if (savedStyles) {
    for (var style in savedStyles) {
      element.style[style] = savedStyles[style];
    }
    delete findInPage.savedElementStyles[element];
  }
};

/**
 * Normalize coordinates according to the current document dimensions. Don't go
 * too far off the screen in either direction. Try to center if possible.
 * @param {Element} elem Element to find normalized coordinates for.
 * @return {Array.<number>} Normalized coordinates.
 */
findInPage.getNormalizedCoordinates = function(elem) {
  var pos = findInPage.findAbsoluteCenter(elem);
  var maxX = Math.max(findInPage.getBodyWidth(), pos[0] + elem.offsetWidth / 2);
  var maxY = Math.max(findInPage.getBodyHeight(),
                      pos[1] + elem.offsetHeight / 2);
  // Don't go too far off the screen in either direction.  Try to center if
  // possible.
  var xPos = Math.max(0,
      Math.min(maxX - window.innerWidth,
               pos[0] - (window.innerWidth / 2)));
  var yPos = Math.max(0,
      Math.min(maxY - window.innerHeight,
               pos[1] - (window.innerHeight / 2)));
  return [xPos, yPos];
};

/**
 * Scale coordinates according to the width of the screen, in case the screen
 * is zoomed out.
 * @param {Array.<number>} coordinates Coordinates to scale.
 * @return {Array.<number>} Scaled coordinates.
 */
findInPage.scaleCoordinates = function(coordinates) {
  var scaleFactor = findInPage.getPageScale();
  return [coordinates[0] * scaleFactor, coordinates[1] * scaleFactor];
};

/**
 * Finds the position of the result and scrolls to it.
 * @param {number} opt_index Index to replace findInPage.index with.
 * @return {string} JSON encoded array of the scroll coordinates "[x, y]".
 */
findInPage.findScrollDimensions = function(opt_index) {
  if (opt_index !== undefined) {
    findInPage.index = opt_index;
  }
  var elem = findInPage.spans[findInPage.index];
  if (!elem) {
    return '';
  }
  var normalized = findInPage.getNormalizedCoordinates(elem);
  var xPos = normalized[0];
  var yPos = normalized[1];

  // Perform the scroll.
  //window.scrollTo(xPos, yPos);

  if (xPos < window.pageXOffset ||
      xPos >= (window.pageXOffset + window.innerWidth) ||
      yPos < window.pageYOffset ||
      yPos >= (window.pageYOffset + window.innerHeight)) {
    // If it's off the screen.  Wait a bit to start the highlight animation so
    // that scrolling can get there first.
    window.setTimeout(findInPage.addHighlightToIndex, 250);
  } else {
    findInPage.addHighlightToIndex();
  }
  var scaled = findInPage.scaleCoordinates(normalized);
  return '[' + scaled.join(',') + ']';
};

/**
 * Initialize the findInPage module.
 */
findInPage.init = function(width, height) {
  if (findInPage.hasInitialized) {
    return;
  }
  findInPage.pageWidth = width;
  findInPage.pageHeight = height;
  findInPage.enable();
  findInPage.hasInitialized = true;
};

/**
 * When the GSA app detects a zoom change, we need to update our css.
 */
findInPage.fixZoom = function(width, height) {
  findInPage.pageWidth = width;
  findInPage.pageHeight = height;
  if (findInPage.style) {
    var selected = document.getElementsByClassName('findysel');
    findInPage.removeStyle();
    for (var i = 0; i < selected.length; i++) {
      var element = selected[0];
      findInPage.removeSelectHighlight(element);
      findInPage.addSelectHighlight(element);
    }
    findInPage.addStyle();
  }
};

/**
 * Enable the findInPage module.  Mainly just adds the style for the classes.
 */
findInPage.enable = function() {
  if (findInPage.style) {
    // Already enabled.
    return;
  }
  findInPage.addStyle();
};

/**
 * Gets the scale ratio between the application window and the web document.
 * @return {number} Scale.
 */
findInPage.getPageScale = function() {
  return (findInPage.pageWidth / findInPage.getBodyWidth());
};

findInPage.getHighlightZoomFactor = function() {
  return Math.min(findInPage.MAX_SELECTION_ZOOM, 1 / findInPage.getPageScale());
};

/**
 * Maximum padding added to a highlighted item when selected.
 * @type {number}
 */
findInPage.MAX_HIGHLIGHT_PADDING = 10;

/**
 * Adds the appropriate style element to the page.
 */
findInPage.addStyle = function() {
  var styleContent = [];
  function addCSSRule(name, style) {
    styleContent.push(name, '{', style, '}');
  };
  var zoom = findInPage.getHighlightZoomFactor();
  var left = ((1 - 1 / zoom) / 2 * 100);
  addCSSRule('.' + findInPage.CSS_CLASS_NAME,
             'background-color:#ffff00 !important;' +
             'padding:0px;margin:0px;' +
             'overflow:visible !important;');
  addCSSRule('.findysel',
             'background-color:#ff9632 !important;' +
             'padding:0px;margin:0px;' +
             'overflow:visible !important;');
  findInPage.style = document.createElement('style');
  findInPage.style.id = 'findInPageStyle';
  findInPage.style.setAttribute('type', 'text/css');
  findInPage.style.appendChild(document.createTextNode(styleContent.join('')));
  document.body.appendChild(findInPage.style);
};

/**
 * Removes the style element from the page.
 */
findInPage.removeStyle = function() {
  if (findInPage.style) {
    document.body.removeChild(findInPage.style);
    findInPage.style = null;
  }
};

/**
 * Disables the findInPage module.  Basically just removes the style and class
 * names.
 */
findInPage.disable = function() {
  if (findInPage.style) {
    findInPage.removeStyle();
    window.setTimeout(findInPage.clearHighlight, 0);
  }
};

/**
* Returns the width of the document.body.  Sometimes though the body lies to
* try to make the page not break rails, so attempt to find those as well.
* An example: wikipedia pages for the ipad.
*/
findInPage.getBodyWidth = function() {
  // It might be better to always use scrollWidth rather than looking
  // at all of the bounding client rects, but I'm not 100% confident and this
  // is known to work.
  var bodyWidth = Math.max(document.body.offsetWidth,
                           document.body.scrollWidth);
  if (bodyWidth <= window.innerWidth) {
    // Calculate the width as the rightmost edge of the document children.
    var newWidth = 0;
    for (var i = document.body.children.length - 1; i >= 0; i--) {
      newWidth = Math.max(newWidth,
          document.body.children[i].getBoundingClientRect().right +
              window.pageXOffset);
    }

    if (newWidth > bodyWidth) {
      return newWidth;
    }
  }
  return bodyWidth;
};

/**
 * Returns the height of the document.body.  Sometimes though the body lies to
 * try to make the page not break rails, so attempt to find those as well.
 * An example: wikipedia pages for the ipad.
 */
findInPage.getBodyHeight = function() {
  // It might be better to always use scrollHeight rather than looking
  // at all of the bounding client rects, but I'm not 100% confident and this
  // is known to work.
  var bodyHeight = Math.max(document.body.offsetHeight,
                            document.body.scrollHeight);
  if (bodyHeight <= window.innerHeight) {
    // Calculate the height as the bottommost edge of the document children.
    var newHeight = 0;
    for (var i = document.body.children.length - 1; i >= 0; i--) {
      newHeight = Math.max(newHeight,
          document.body.children[i].getBoundingClientRect().bottom +
              window.pageYOffset);
    }

    if (newHeight > bodyHeight) {
      return newHeight;
    }
  }
  return bodyHeight;
};

/**
 * Helper function that determines if an element is visible.
 * @param {Element} elem Element to check.
 * @return {boolean} Whether elem is visible or not.
 */
findInPage.isVisible = function(elem) {
  if (!elem) {
    return false;
  }
  var top = 0;
  var left = 0;
  var bottom = Infinity;
  var right = Infinity;

  var originalElement = elem;
  var nextOffsetParent = originalElement.offsetParent;
  var computedStyle = document.defaultView.getComputedStyle(elem, null);

  // We are currently handling all scrolling through the app, which means we can
  // only scroll the window, not any scrollable containers in the DOM itself. So
  // for now this function returns false if the element is scrolled outside the
  // viewable area of its ancestors.
  // TODO (jonwall): handle scrolling within the DOM.
  var pageHeight = findInPage.getBodyHeight();
  var pageWidth = findInPage.getBodyWidth();

  while (elem && elem.nodeName != 'BODY') {
    if (elem.style.display == 'none' ||
        elem.style.visibility == 'hidden' ||
        computedStyle.display == 'none' ||
        computedStyle.visibility == 'hidden') {
      return false;
    }

    // For the original element and all ancestor offsetParents, trim down the
    // visible area of the original element.
    if (elem.isSameNode(originalElement) || elem.isSameNode(nextOffsetParent)) {
      var visible = elem.getBoundingClientRect();
      if (visible.width == 0 || visible.height == 0) {
        return false;
      }

      // Adjust the visible viewport for the original element and parent
      // offset nodes that clip overflow.
      if (elem.isSameNode(originalElement) ||
          computedStyle.overflow != 'visible') {
        top = Math.max(top, visible.top + window.pageYOffset);
        bottom = Math.min(bottom, visible.bottom + window.pageYOffset);
        left = Math.max(left, visible.left + window.pageXOffset);
        right = Math.min(right, visible.right + window.pageXOffset);
      }

      // The element is not within the original viewport.
      if (top < 0 || left < 0 ||
          // The viewable portion of the element is no longer two dimensional.
          bottom < top || right < left ||
          // The element is flowing off the boundary of the page. Note this is
          // not comparing to the size of the window, but the calculated offset
          // size of the document body. This can happen if the element is within
          // a scrollable container in the page.
          right > pageWidth || bottom > pageHeight) {
        return false;
      }
      nextOffsetParent = elem.offsetParent;
    }

    elem = elem.parentNode;
    computedStyle = document.defaultView.getComputedStyle(elem, null);
  }
  return true;
};

/**
 * Helper function to find the absolute center of an element on the page.
 * @param {Element} elem Element to check.
 * @return {Array.<number>} [x, y] positions.
 */
findInPage.findAbsoluteCenter = function(elem) {
  var boundingRect = elem.getBoundingClientRect();
  return [boundingRect.right - boundingRect.width / 2 + window.pageXOffset,
    boundingRect.bottom - boundingRect.height / 2 + window.pageYOffset];
};

/**
 * Helper function to find the absolute position of an element on the page.
 * @param {Element} elem Element to check.
 * @return {Array.<number>} [x, y] positions.
 */
findInPage.findAbsolutePosition = function(elem) {
  var boundingRect = elem.getBoundingClientRect();
  return [boundingRect.left + window.pageXOffset,
      boundingRect.top + window.pageYOffset];
};

/**
 * HTML escapes the text.
 * @param {string} text Text to escape.
 * @param {string} escaped text.
 */
findInPage.escapeHTML = function(text) {
  var unusedDiv = document.createElement('div');
  unusedDiv.innerText = text;
  return unusedDiv.innerHTML;
};

/**
 * Escapes regexp special characters.
 * @param {string} text Text to escape.
 * @return {string} escaped text.
 */
findInPage.escapeRegex = function(text) {
  return text.replace(findInPage.REGEX_ESCAPER, '\\$1');
};

